#!/usr/bin/env python3

import sys
import os
import re
import pprint
import subprocess
import shutil

def params_to_env(params):
    for item in params:
        os.environ[item]=params[item]
    f=open("d8_objects.txt","w")
    f.write(params['D8_OBJECTS'])
    f.close()

def patch_paths(params,s):

    def repl_path(m):
        return m.group(1)+params['V8_COMPILEDIR']+'/'+m.group(2)

    s=re.sub('\.\./\.\.',params['V8_BASEDIR'],s)
    s=re.sub('(-I)(gen)',repl_path,s)
    return s

def remove_unwanted(params,s):
    s=re.sub(' -Werror','',s)
    s=re.sub(' -fno-exceptions','',s)
    s=re.sub(' -Wunused-variable','',s)
    #s=re.sub(' -Wno-unused-parameter','',s)
    s=re.sub(' -DV8_DEPRECATION_WARNINGS','',s)
    s=re.sub(' -Wl,-z,defs','',s)
    #s=re.sub(' -fvisibility=default',' -fvisibility=public',s)
    return s

def array_unique(arr):
    ret=[]
    used={}
    for item in arr:
        if not (item in used):
            used[item]=1
            ret.append(item)
    return ret

def parse_d8_compile_args(params):
    f=open(params['V8_COMPILEDIR']+"/obj/d8.ninja")
    build_section=""
    for item in f.readlines():
        tmp=re.search('^( *)(\w+) = (.*)',item)
        if tmp:
            if tmp.group(1)=="" or (tmp.group(1)=="  " and build_section=="./d8"):
                params['D8_'+tmp.group(2).upper()]=remove_unwanted(params,patch_paths(params,tmp.group(3)))
        tmp=re.search('^build ([^ ]+): link (.*) \|\|',item)
        if tmp:
            build_section=tmp.group(1)
            if build_section=="./d8":
                files_objects =[]
                files_archives=[]
                files_libpaths=[]
                for file in tmp.group(2).split(" "):
                    if file.find("/d8/")==-1:
                        if file.endswith(".o"):
                            files_objects .append(params['V8_COMPILEDIR']+"/"+file)
                        elif file.endswith(".a"):
                            tmp=re.search('^(.*)/lib(\w+)\.a$',file)
                            if tmp:
                                files_libpaths.append("-L"+params['V8_COMPILEDIR']+"/"+tmp.group(1))
                                files_archives.append("-l"+tmp.group(2))
                            else:
                                print("Could not parse library name - "+file)
                                exit(1)
                        else:
                            print("Unknown file extension - "+file)
                            exit(1)
                            #files_archives.append(params['V8_COMPILEDIR']+"/"+file)
                files_objects .append(params['V8_COMPILEDIR']+"/obj/v8_libbase/stack_trace.o") # ???
                files_objects .append(params['V8_COMPILEDIR']+"/obj/v8_libbase/stack_trace_posix.o") # ???
                files_archives.append(params['V8_COMPILEDIR']+"/obj/libwee8.a") # ???
                #files_objects .append(params['V8_COMPILEDIR']+"/obj/v8_libbase/sys-info.o") # ???
                params['D8_OBJECTS' ]=" ".join(array_unique(files_objects))
                params['D8_ARCHIVES']=" ".join(array_unique(files_archives))
                params['D8_LIBPATHS']=" ".join(array_unique(files_libpaths))
    f.close()

def resolve_3rd_party(params):
    if not os.path.exists("3rd-party"):
        os.mkdir("3rd-party")
    # TODO
    dependencies=[
        {
            "name":"fcgi",
            "includes":["fcgiapp.h","fcgi_config.h","fcgimisc.h","fcgio.h","fcgios.h","fcgi_stdio.h"],
            "libraries":["fcgi"]
        },
        {
            "name":"gd",
            "includes":["gdfx.h","gd.h","gd_io.h"],
            "libraries":["gd"]
        }
    ]
    for dep in dependencies:
        for file in dep["includes"]:
            if os.path.exists("/usr/include/"+file):
                shutil.copyfile("/usr/include/"+file,"3rd-party/"+file)
            else:
                print("File %{file}s not found in /usr/include/ - TODO",{file:file})
                exit(1)

        arr=[]
        for file in dep["libraries"]:
            arr.append("-l"+file)
        params[dep["name"].upper()+'_LIBRARY']=" ".join(arr)

def work():
    if len(sys.argv)==1:
        print("Usage:")
        myname=sys.argv[0]
        repl= {'myname':myname}
        print("    %(myname)s build                            - builds teajs" % repl)
        print("    %(myname)s build --v8 some_v8_path/         - builds teajs assuming another v8 location" % repl)
        print("    sudo %(myname)s install                     - installs teajs" % repl)
        print("    %(myname)s remoteinstall 172.168.32.1       - remote installs teajs (connects root@172.168.32.1)" % repl)
        print("    %(myname)s clean                            - cleans project" % repl)
        return 1

    # TODO --args

    params={}
    params['CDIR']=os.path.dirname(os.path.realpath(__file__))
    params['PDIR']=re.sub('([\\\/])[^\\\/]+[\\\/]?$','',params['CDIR'])
    #params['PDIR']=os.path.join(os.path.split(params['CDIR']),"..")
    params['V8_BASEDIR']=params['PDIR']+"/v8_things/v8"
    params['V8_COMPILEDIR']=params['V8_BASEDIR']+"/out/x64.release";

    params['TEAJS_BASEDIR']=params['CDIR']
    params['TEAJS_LIBPATH']=params['TEAJS_BASEDIR']+"/lib"
    f=open("VERSION")
    params['TEAJS_VERSION']=f.read().split("\n")[0]
    f.close()
    params['V8_CPP']=params['V8_BASEDIR']+"/third_party/llvm-build/Release+Asserts/bin/clang++"

    if os.path.exists("/usr/bin/pkg-config"):
        params['LIBPQ_INCLUDE']=subprocess.check_output('pkg-config --cflags libpq', shell=True).decode("utf-8").split("\n")[0]
        params['LIBPQ_LIBRARY']=subprocess.check_output('pkg-config --libs libpq', shell=True).decode("utf-8").split("\n")[0]
    else:
        print("ERROR: NO /usr/bin/pkg-config - TODO!!!")
        return 1

    resolve_3rd_party(params)
    # TODO
    #params['FCGI_INCLUDE']="-I/usr/include"
    params['JOBS']="8"

    parse_d8_compile_args(params)
    #export STD_CPP=`gcc -v --help 2>&1 | grep "std=.*ISO 2020" | head -1 | sed 's/ *-std=//' |sed 's/ .*//'`

    #params['COMPILE_FLAGS']=`cat linker_info.txt | grep 1/3 | sed 's/.*shell.o.d //' | sed 's/ -c .*//'|sed "s+\\.\\./\\.\\.+$V8_BASEDIR+g"|sed 's/ -fno-exceptions//'|sed 's/ -Werror//'|sed 's/ -DV8_DEPRECATION_WARNINGS//'`
    #pprint.pprint(params)
    f=open("build-tea.log","w")
    for item in params:
        print("    "+item+" = "+params[item][:70])
        f.write(item+" = "+params[item]+"\n")
    f.close()

    if sys.argv[1]=="build":
        #os.remove("tea")
        #os.remove("tea.exe")
        #os.remove("libtea.exe")
        #for item in os.listdir(params['TEAJS_LIBPATH']):
        #    if item.endswith('.dll') or item.endswith('.so') or item.endswith('.dylib'):
        #        os.remove(params['TEAJS_LIBPATH']+"/"+item)
        params_to_env(params)
        subprocess.call(["make","-j",params["JOBS"]])
    elif sys.argv[1]=="install":
        params_to_env(params)
        subprocess.call(["make","-j",params["JOBS"],"install"])
    elif sys.argv[1]=="remoteinstall":
        params['REMOTE']=sys.argv[2]
        params_to_env(params)
        subprocess.call(["make","-j",params["JOBS"],"remoteinstall"])
    elif sys.argv[1]=="clean":
        params_to_env(params)
        subprocess.call(["make","-j",params["JOBS"],"clean"])


if __name__ == "__main__":
    work()


